class Types:
    TEXT = 'text'
    DECIMAL = 'decimal'
    DATE = 'date'
    DATE_TIME = 'date_time'
    BOOLEAN = 'boolean'

    LIST_OF_DECIMALS = 'list(integer,decimal)'
    LIST_OF_TEXTS = 'list(integer,text)'

    TIME_SERIES_OF_DECIMALS = 'timeseries(date,decimal)'
